#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <stdio.h>
#include <stdlib.h>

#include <errno.h>

#include <sys/inotify.h>
#include <limits.h>

#define checkError(expr,userMsg) \
    do { \
        if (!(expr)) {\
            perror(userMsg); \
            exit(EXIT_FAILURE); \
        } \
    } while (0)

#define NAME_MAX (255)
#define MAX_SIZE (10*sizeof(struct inotify_event)*NAME_MAX + 1)

void displayEvents(struct inotify_event* event) {

    printf("wd: %d\n", event->wd);
    if (event->cookie > 0) {
        printf("cookie: %d\n", event->cookie);
    }

    printf("Mask: \n");
    if (event->mask & IN_ACCESS)
        printf("IN_ACCESS ");
    if (event->mask & IN_ATTRIB)
        printf("IN_ATTRIB ");
    if (event->mask & IN_CLOSE_WRITE)
        printf("IN_CLOSE_WRITE ");
    if (event->mask & IN_CLOSE_NOWRITE)
        printf("IN_CLOSE_NOWRITE ");
    if (event->mask & IN_CREATE)
        printf("IN_CREATE ");
    if (event->mask & IN_DELETE)
        printf("IN_DELETE ");
    if (event->mask & IN_DELETE_SELF)
        printf("IN_DELETE_SELF ");
    if (event->mask & IN_IGNORED)
        printf("IN_IGNORED ");
    if (event->mask & IN_ISDIR)
        printf("IN_ISDIR ");
    if (event->mask & IN_MODIFY)
        printf("IN_MODIFY ");
    if (event->mask & IN_MOVE_SELF)
        printf("IN_MOVE_SELF ");
    if (event->mask & IN_MOVED_TO)
        printf("IN_MOVED_TO ");
    if (event->mask & IN_MOVED_FROM)
        printf("IN_MOVED_FROM ");
    if (event->mask & IN_OPEN)
        printf("IN_OPEN ");
    if (event->mask & IN_UNMOUNT)
        printf("IN_UNMOUNT ");
    if (event->mask & IN_Q_OVERFLOW)
        printf("IN_Q_OVERFLOW ");
    printf("\n");

    if (event->len > 0) {
        printf("Name: %s\n", event->name);
    }
}

void process_all_events(char* buf, int readBytes) {

    char* p = NULL;

    for (p = buf; p < buf + readBytes;  ) {

        struct inotify_event* event = (struct inotify_event*)p;
        displayEvents(event);

        p += sizeof(struct inotify_event)*event->len;
    }
}

int main(int argc, char** argv) {

    errno = EINVAL;
    checkError(argc > 1, "args");
    errno = 0;

    int inotifyFd = inotify_init();
    checkError(inotifyFd != -1, "inotify_init");

    for (int i = 1; i < argc; i++) {
        int wd = inotify_add_watch(inotifyFd, argv[i], IN_ALL_EVENTS);
        checkError(wd != -1, "add watch");

        printf("nadgledam file %s sa watch %d\n", argv[i], wd);
    }

    char buffer[MAX_SIZE];
    while (1) {
        int readBytes = read(inotifyFd, buffer, MAX_SIZE);
        checkError(readBytes != -1, "read");
        if (readBytes == 0) {
            printf("No events");
            break;
        }

        printf("Procitao: %dB", readBytes);

        process_all_events(buffer, readBytes);
    }

    close(inotifyFd);

    exit(EXIT_SUCCESS);
}